/*
 * CS 106X Fall 2018 - Nick Troccoli
 * C++ solutions for problems from Section #9 (Week 10)
 * Section topic: Hashing, sorting, and review!
 */

#include <iostream>
#include <fstream>
#include "console.h"
#include "grid.h"
#include "set.h"
#include "basicgraph.h"
#include "queue.h"
#include "section9-solutions.h"

using namespace std;

/*****************************************************************
 * 1-2. Hash function quality and HashMap simulation - (hashing) *
 *****************************************************************/

// no code required for these problems; see solution handout

/**********************************
 * 3. Mergesort trace - (sorting) *
 **********************************/

// no code required for this problem; see solution handout


/*******************************
 * 4. consume - (linked lists) *
 *******************************/
void consume(ListNode*& list1, ListNode*& list2) {
    if (list1 == nullptr) {
        list1 = list2;
    } else {
        ListNode* current = list1;
        while (current->next != nullptr) {
            current = current->next;
        }
        current->next = list2;
    }
    list2 = nullptr;
}


/*************************************
 * 5. transferEvens - (linked lists) *
 *************************************/
ListNode* transferEvens(ListNode*& list1) {
    ListNode* list2 = nullptr;
    if (list1 != nullptr) {
        list2 = list1;
        list1 = list1->next;
        ListNode* current = list1;
        ListNode* list2Last = list2;
        while (current != nullptr && current->next != nullptr) {
            list2Last->next = current->next;
            list2Last = current->next;
            current->next = current->next->next;
            current = current->next;
        }
        list2Last->next = nullptr;
    }
    return list2;
}


/****************************************
 * 6. permutations - (recursion warmup) *
 ****************************************/
void permutationsHelper(string s, string chosen) {
    if (s.size() == 0) {
        cout << chosen << " ";   // base case: no choices left to be made
    } else {
        // recursive case: choose each possible next letter
        for (int i = 0; i < s.size(); i++) {
            char ch = s[i];                         // choose
            s = s.substr(0, i) + s.substr(i + 1);
            permutationsHelper(s, chosen + ch);     // explore
            s = s.substr(0, i) + ch + s.substr(i);  // un-choose
        }
    }
}

void permutations(string s) {
    permutationsHelper(s, "");
}


/******************************************
 * 7. Down on the corner - (backtracking) *
 ******************************************/
bool canMoveToCornerHelper(int r, int c, const Grid<int>& grid, Grid<bool>& visited) {
    /* If we are out of bounds or been here before, we can't continue */
    if(!grid.inBounds(r, c) || visited[r][c]) return false;
    /* If we are at the final spot, we've made it! */
    if(r == grid.numRows() - 1 && c == grid.numCols() - 1) return true;
    /* Mark this spot as visited so we don't end up in an infinite loop. */
    visited[r][c] = true;

    int currNum = grid[r][c];           // how much we can move

    /* Try moving in each direction by the allowed amount */
    return canMoveToCornerHelper(r + currNum, c, grid, visited)
            || canMoveToCornerHelper(r - currNum, c, grid, visited)
            || canMoveToCornerHelper(r, c + currNum, grid, visited)
            || canMoveToCornerHelper(r, c - currNum, grid, visited);
}

bool canMoveToCorner(const Grid<int> &grid) {
    /* Need to keep track of places we've visited to avoid
     * infinite recursion. */
    Grid<bool> visited(grid.numRows(), grid.numCols(), false);
    /* 0,0 is upper left corner */
    return canMoveToCornerHelper(0, 0, grid, visited);
}


/************************************
 * 8. limitPathSum - (binary trees) *
 ************************************/
void deleteTree(TreeNode*& node) {
    if (node == nullptr) return;
    deleteTree(node->left);
    deleteTree(node->right);
    delete node;
}

void limitPathSumHelper(TreeNode*& node, int max, int sum) {
    if (node != nullptr) {
        sum += node->data;
        if (sum > max) {
            deleteTree(node);
            node = nullptr;
        } else {
            limitPathSumHelper(node->left,  max, sum);
            limitPathSumHelper(node->right, max, sum);
        }
    }
}

void limitPathSum(TreeNode*& node, int max) {
    limitPathSumHelper(node, max, 0);
}


/**************************************************************************************
 * 9. The great tree list recursion problem - (binary trees, linked lists, recursion) *
 **************************************************************************************/
TreeNode* lastNode(TreeNode* list) {
    if(list == nullptr) return nullptr;
    while(list->right != nullptr) {
        list = list->right;
    }
    return list;
}

TreeNode* flattenTree(TreeNode* root) {
    /* If root is null we already have a DLL :) */
    if(root == nullptr) return nullptr;

    TreeNode* head = root;      // this will be the head of the DLL we return

    /* Recursively convert left and right subtree to a DLL, where
     * the value returned by the recursion is the **HEAD** of
     * the converted DLL.
     */
    TreeNode* left = flattenTree(root->left);
    TreeNode* right = flattenTree(root->right);

    /* We want to combine the left DLL, current node, and right DLL.
     * To do this we need the last node in the left DLL so we can
     * connect it to the current node.
     */
    TreeNode* leftLast = lastNode(left);
    root->left = leftLast;

    /* If the left list isn't empty, then we should set the head of
     * the DLL we are going to return to be the head of the left DLL.
     */
    if(leftLast != nullptr) {
        head = left;
        leftLast->right = root;
    }

    /* Connect current node to the right DLL */
    root->right = right;

    if(right != nullptr) {
        right->left = root;
    }

    /* Return head of new DLL */
    return head;
}


/******************************
 * 10. isBipartite - (graphs) *
 ******************************/
bool isBipartite(const BasicGraph& graph) {
    Set<string> red, blue;         // to store two bipartite sets (one red, one blue)
    Map<Vertex*, string> colorMap; // to track which set vertices are in, if any
    Queue<Vertex*> queue;          // queue to visit vertices in a breadth-first search

    Vertex* start = graph.getVertexSet().first();   // any vertex as starting point
    red.add(start->name);
    colorMap[start] = "red";
    queue.enqueue(start);

    while (!queue.isEmpty()) {
        Vertex* v = queue.dequeue();
        for (Vertex* neighbor : graph.getNeighbors(v)) {
            if (!colorMap.containsKey(neighbor)) {
                // give neighbor the opposite color of v
                if (colorMap[v] == "red") {
                    blue.add(neighbor->name);
                    colorMap[neighbor] = "blue";
                } else {
                    red.add(neighbor->name);
                    colorMap[neighbor] = "red";
                }
                queue.enqueue(neighbor);
            } else if (colorMap[v] == colorMap[neighbor]) {
                // neighbors of same color; cannot be bipartite
                return false;
            } // else they are of opposite color, which is fine
        }
    }

    cout << "set 1: " << red  << endl;  // if we get here, partition was successful
    cout << "set 2: " << blue << endl;
    return true;
}



/* main function for testing individual problems */
int main() {
    cout << "Problem 4 (consume):" << endl;
    ListNode* list1 = buildLinkedList({1, 3, 5, 7});
    ListNode* list2 = buildLinkedList({2, 4, 6});
    consume(list1, list2);
    cout << "consume(list1, list2):" << endl;
    cout << "list1: " << linkedListToString(list1) << endl;
    cout << "list2: " << linkedListToString(list2) << endl;
    list1 = buildLinkedList({1, 3, 5, 7});
    list2 = buildLinkedList({2, 4, 6});
    consume(list2, list1);
    cout << "consume(list2, list1)" << endl;
    cout << "list1: " << linkedListToString(list1) << endl;
    cout << "list2: " << linkedListToString(list2) << endl;


    cout << endl << "Problem 5 (transferEvens):" << endl;
    list1 = buildLinkedList({3, 1, 4, 15, 9, 2, 6, 5, 35, 89});
    list2 = transferEvens(list1);
    cout << "list1: " << linkedListToString(list1) << endl;
    cout << "list2: " << linkedListToString(list2) << endl;


    cout << endl << "Problem 6 (permutations):" << endl;
    permutations("abc");
    cout << endl;

    cout << endl << "Problem 7 (Down on the corner):" << endl;
    Grid<int> grid = {
        {2, 4, 1},
        {3, 2, 1},
        {4, 2, 3},
        {2, 1, 9}
    };
    cout << boolToString(canMoveToCorner(grid)) << endl;

    cout << endl << "Problem 8 (limitPathSum):" << endl;
    TreeNode* root8 = new TreeNode(29,
                                   new TreeNode(17,
                                                new TreeNode(-7,
                                                             new TreeNode(11,
                                                                          nullptr,
                                                                          nullptr
                                                                          ),
                                                             new TreeNode(12,
                                                                          nullptr,
                                                                          nullptr
                                                                          )
                                                             ),
                                                new TreeNode(37,
                                                             nullptr,
                                                             new TreeNode(16,
                                                                          nullptr,
                                                                          nullptr
                                                                          )
                                                             )
                                                ),
                                   new TreeNode(15,
                                                new TreeNode(4,
                                                             nullptr,
                                                             nullptr
                                                             ),
                                                new TreeNode(14,
                                                             new TreeNode(-9,
                                                                          nullptr,
                                                                          nullptr
                                                                          ),
                                                             new TreeNode(19,
                                                                          nullptr,
                                                                          nullptr)
                                                             )
                                                )
                                   );

    cout << "Before call:" << endl;
    printSideways(root8);
    limitPathSum(root8, 50);
    cout << "After call:" << endl;
    printSideways(root8);

    cout << endl << "Problem 9 (The great tree list recursion problem):" << endl;
    TreeNode* root9 = new TreeNode(4,
                                   new TreeNode(2,
                                                new TreeNode(1,
                                                             nullptr,
                                                             nullptr
                                                             ),
                                                new TreeNode(3,
                                                             nullptr,
                                                             nullptr
                                                             )
                                                ),
                                   new TreeNode(6,
                                                new TreeNode(5,
                                                             nullptr,
                                                             nullptr
                                                             ),
                                                nullptr
                                                )
                                   );
    cout << "Before call:" << endl;
    printSideways(root9);
    TreeNode* head9 = flattenTree(root9);
    cout << "After call:" << endl;
    TreeNode* curr = head9;
    cout << curr->data;
    while (curr->right != nullptr) {
        curr = curr->right;
        cout << " <--> " << curr->data;
    }
    cout << endl << "and in reverse: " << endl;
    cout << curr->data;
    while (curr->left != nullptr) {
        curr = curr->left;
        cout << " <--> " << curr->data;
    }
    cout << endl;

    cout << endl << "Problem 10 (isBipartite):" << endl;
    BasicGraph g;
    g.addVertex("A");
    g.addVertex("B");
    g.addVertex("C");
    g.addVertex("D");
    g.addVertex("E");
    g.addVertex("F");
    g.addVertex("G");
    g.addEdge("A", "B", false);
    g.addEdge("A", "D", false);
    g.addEdge("B", "C", false);
    g.addEdge("C", "E", false);
    g.addEdge("D", "F", false);
    g.addEdge("E", "F", false);
    g.addEdge("F", "G", false);
    cout << boolToString(isBipartite(g));

    return 0;
}
