#pragma once

#include "vector.h"

struct ListNode {
    int data;
    ListNode* next;

    ListNode(int data, ListNode* next) {
        this->data = data;
        this->next = next;
    }

    ListNode(int data) {
        this->data = data;
        this->next = nullptr;
    }
};

/* Function to conveniently construct a linked list from a data vector */
ListNode* buildLinkedList(const Vector<int>& list) {
    if(list.isEmpty()) {
        return nullptr;
    }
    ListNode* front = new ListNode(list[0], nullptr);
    ListNode* current = front;
    for (int i = 1; i < list.size(); i++) {
        current->next = new ListNode(list[i], nullptr);
        current = current->next;
    }
    return front;
}

std::string linkedListToString(const ListNode* front) {
    std::stringstream toReturn;
    toReturn << "{";
    if (front != nullptr) {
        toReturn << front->data;
        front = front->next;
        while (front != nullptr) {
            toReturn << ", " << front->data;
            front = front->next;
        }
    }
    toReturn << "}";
    return toReturn.str();
}

struct TreeNode {
    int data;
    TreeNode* left;
    TreeNode* right;

    TreeNode(int data, TreeNode* left, TreeNode* right) {
        this->data = data;
        this->left = left;
        this->right = right;
    }

    TreeNode(int data) {
        this->data = data;
        this->left = nullptr;
        this->right = nullptr;
    }
};

/*
 * Function to visualize a tree rooted at the provided node.
 * The tree is printed 90 degrees to the left, so right nodes are towards
 * the top, and left nodes are towards the bottom of the console window.
 */
template <typename NodeType>
void printSidewaysHelper(NodeType* node, std::string indent) {
    if (node) {
        printSidewaysHelper<NodeType>(node->right, indent + "    ");
        std::cout << indent;
        std::cout << node->data;
        std::cout << std::endl;
        printSidewaysHelper<NodeType>(node->left, indent + "    ");
    }
}

template <typename NodeType>
void printSideways(NodeType* node, std::string indent = "") {
    if (node == nullptr) {
        std::cout << "(empty tree)" << std::endl;
    } else {
        std::cout << std::endl << "(right)" << std::endl;
        printSidewaysHelper<NodeType>(node, indent);
        std::cout << "(left)" << std::endl << std::endl;
    }
}
